"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ast_modifier_1 = require("./ast-modifier/ast-modifier");
const autofix_fk_1 = require("./ast-modifier/modifiers/autofix-fk");
const ConsistencyChecker_1 = require("./consistency-checker/ConsistencyChecker");
const dialect_generators_1 = require("./dialect-generators");
const parse_ddl_1 = require("./parser/parse-ddl");
const rds_provider_1 = require("./rds-provider/rds-provider");
function generateDdl(ast, dialect, options) {
    return dialect_generators_1.generateDdl(ast, dialect, options);
}
exports.generateDdl = generateDdl;
function parseDdl(source, options = {}) {
    let ast = parse_ddl_1.parseDdlToAst(source);
    if (options.autofix)
        ast = autofix(ast, options.autofix === true ? undefined : options.autofix);
    if (options.checkConsistency) {
        const report = checkConsistency(ast);
        if (!report.valid)
            throw new Error(report.errors.join("\n"));
    }
    return options.freeze ? deepFreezePojo(ast) : ast;
}
exports.parseDdl = parseDdl;
function autofix(input, options) {
    const modifiers = [];
    const afterEndListeners = [];
    if (!options || options.foreignKeys) {
        const { astModifiers, afterEnd } = autofix_fk_1.autofixFk();
        modifiers.push(...astModifiers);
        if (afterEnd)
            afterEndListeners.push(afterEnd);
    }
    if (modifiers.length === 0)
        return input;
    const output = ast_modifier_1.modifyAst(input, modifiers);
    for (const listener of afterEndListeners)
        listener();
    return output;
}
exports.autofix = autofix;
function checkConsistency(ast) {
    const checker = new ConsistencyChecker_1.default();
    checker.check(ast);
    return checker.getConsistencyCheckerReport();
}
exports.checkConsistency = checkConsistency;
function parseDdlToRds(source, options = {}) {
    const ast = parseDdl(source, {
        autofix: options.autofix,
        freeze: false,
        checkConsistency: true
    });
    return createRds(ast, { freeze: !!options.freeze });
}
exports.parseDdlToRds = parseDdlToRds;
function createRds(ast, options = {}) {
    if (options.checkConsistency) {
        const report = checkConsistency(ast);
        if (!report.valid)
            throw new Error(report.errors.join("\n"));
    }
    const rds = rds_provider_1.createRdsFromAst(ast);
    return options.freeze ? deepFreezePojo(rds) : rds;
}
exports.createRds = createRds;
function deepFreezePojo(object) {
    if (Object.isFrozen(object))
        return object;
    Object.freeze(object);
    for (const key of Object.keys(object)) {
        const value = object[key];
        if (value && typeof value === "object")
            deepFreezePojo(value);
    }
    return object;
}
