"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ast_walker_1 = require("./ast-walker");
function isAstListener(modifier) {
    return modifier["forEach"] !== undefined && !!modifier["listen"];
}
function isAstReplacer(modifier) {
    return modifier["forEach"] !== undefined && !!modifier["replace"];
}
function isAstInserter(modifier) {
    return modifier["forEachChildOf"] !== undefined;
}
function modifyAst(ast, modifiers) {
    const exec = new AstModifierExecutor(modifiers);
    return exec.modifyAst(ast);
}
exports.modifyAst = modifyAst;
class AstModifierExecutor {
    constructor(modifiers) {
        this.listeners = normalizeAstListeners(modifiers.filter(isAstListener));
        this.replacers = normalizeAstReplacers(modifiers.filter(isAstReplacer));
        this.inserters = normalizeAstInserters(modifiers.filter(isAstInserter));
    }
    modifyAst(node) {
        return this.walk(node, ast_walker_1.astWalker) || { orders: [] };
    }
    walk(node, walker) {
        if (ast_walker_1.isObjectWalker(walker)) {
            if (walker.self) {
                return applyReplacersOnSelf(node, this.getHookListener(walker), this.getHookReplacer(walker), node => this.walk(node, walker.self));
            }
            return applyReplacersOnObject(node, this.getHookListener(walker), this.getHookReplacer(walker), this.getReplacersOfChildren(walker));
        }
        if (ast_walker_1.isArrayWalker(walker)) {
            return applyReplacersOnArray(node, this.getHookListener(walker), this.getHookReplacer(walker), this.getReplacerOfChild(walker), this.getHookInserterForChild(walker));
        }
        const childWalker = walker(node);
        return childWalker ? this.walk(node, childWalker) : node;
    }
    getHookListener({ hookName }) {
        return (hookName && this.listeners.get(hookName)) || emptyListener;
    }
    getHookReplacer({ hookName }) {
        return (hookName && this.replacers.get(hookName)) || keepSame;
    }
    getHookInserterForChild({ hookName }) {
        if (hookName)
            return this.inserters.get(hookName);
    }
    getReplacersOfChildren({ children }) {
        if (!children)
            return {};
        const childModifiers = {};
        for (const [childName, childWalker] of Object.entries(children))
            childModifiers[childName] = childNode => this.walk(childNode, childWalker);
        return childModifiers;
    }
    getReplacerOfChild({ child }) {
        if (!child)
            return;
        return childNode => this.walk(childNode, child);
    }
}
function keepSame(something) {
    return something;
}
// tslint:disable-next-line: no-empty
function emptyListener() {
}
function applyReplacersOnSelf(source, listener, replacer, selfReplacer) {
    listener(source);
    const updated = replacer(source);
    if (updated)
        return selfReplacer(updated);
}
function applyReplacersOnObject(source, listener, replacer, childReplacers) {
    listener(source);
    let copy;
    const updated = replacer(source);
    if (updated !== undefined) {
        if (updated !== source)
            copy = updated;
        for (const [childName, child] of Object.entries(updated)) {
            const modifier = childReplacers[childName];
            if (!modifier)
                continue;
            const updatedChild = modifier(child);
            if (updatedChild !== child) {
                if (!copy)
                    copy = Object.assign({}, updated);
                if (updatedChild === undefined)
                    delete copy[childName];
                else
                    copy[childName] = updatedChild;
            }
        }
    }
    return copy || updated;
}
function applyReplacersOnArray(source, listener, replacer, childReplacer, childInserter) {
    listener(source);
    let copy;
    const updated = replacer(source);
    if (updated !== undefined && (childReplacer || childInserter)) {
        if (updated !== source)
            copy = updated;
        const insertedAfter = [];
        for (const [index, child] of updated.entries()) {
            let updatedChild = child;
            if (childReplacer) {
                updatedChild = childReplacer(child);
                if (updatedChild !== child) {
                    if (!copy)
                        copy = [...updated];
                    copy[index] = updatedChild;
                }
            }
            if (childInserter && updatedChild !== undefined) {
                let newChildren = childInserter(updatedChild);
                if (newChildren && childReplacer)
                    newChildren = newChildren.map(childReplacer).filter(child => child !== undefined);
                if (newChildren)
                    insertedAfter.push({ index: index + 1, newChildren });
            }
        }
        if (insertedAfter.length > 0) {
            if (!copy)
                copy = [...source];
            let insertCount = 0;
            for (const { index, newChildren } of insertedAfter) {
                copy.splice(index + insertCount, 0, ...newChildren);
                insertCount += newChildren.length;
            }
        }
        if (copy)
            copy = copy.filter(child => child !== undefined);
    }
    return copy || source;
}
function normalizeAstListeners(replacers) {
    const map = new Map();
    for (const modifier of replacers) {
        let list = map.get(modifier.forEach);
        if (!list) {
            list = [];
            map.set(modifier.forEach, list);
        }
        list.push(modifier.listen);
    }
    const mapOfOne = new Map();
    for (const [hookName, list] of map.entries())
        mapOfOne.set(hookName, toOneListener(list));
    return mapOfOne;
}
function toOneListener(replacers) {
    return node => {
        for (const replacer of replacers)
            replacer(node);
    };
}
function normalizeAstReplacers(replacers) {
    const map = new Map();
    for (const modifier of replacers) {
        let list = map.get(modifier.forEach);
        if (!list) {
            list = [];
            map.set(modifier.forEach, list);
        }
        list.push(modifier.replace);
    }
    const mapOfOne = new Map();
    for (const [hookName, list] of map.entries())
        mapOfOne.set(hookName, toOneReplacer(list));
    return mapOfOne;
}
function toOneReplacer(replacers) {
    return node => {
        for (const replacer of replacers) {
            node = replacer(node);
            if (node === undefined)
                return;
        }
        return node;
    };
}
function normalizeAstInserters(inserters) {
    const map = new Map();
    for (const modifier of inserters) {
        let list = map.get(modifier.forEachChildOf);
        if (!list) {
            list = [];
            map.set(modifier.forEachChildOf, list);
        }
        list.push(modifier.insertAfter);
    }
    const mapOfOne = new Map();
    for (const [hookName, list] of map.entries())
        mapOfOne.set(hookName, toOneInserter(list));
    return mapOfOne;
}
function toOneInserter(inserters) {
    return node => {
        const merged = [];
        for (const inserter of inserters) {
            const insert = inserter(node);
            if (insert !== undefined)
                merged.push(...insert);
        }
        if (merged.length > 0)
            return merged;
    };
}
