"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const modifier_helpers_1 = require("../modifier-helpers");
const fk_column_to_table_constraint_1 = require("./fk-column-to-table-constraint");
function autofixFk() {
    const knownTables = new Set();
    const deferredFk = new Map();
    let currentTableName;
    function addAlterTable(fk) {
        if (!currentTableName)
            throw new Error(`Missing current table name!`);
        let list = deferredFk.get(fk.referencedTable);
        if (!list) {
            list = [];
            deferredFk.set(fk.referencedTable, list);
        }
        list.push({
            orderType: "alterTable",
            table: currentTableName,
            add: [fk]
        });
    }
    const astModifiers = [
        {
            forEach: "createTable",
            listen(node) {
                knownTables.add(node.name);
                currentTableName = node.name;
            }
        },
        {
            forEach: "alterTable",
            listen(node) {
                currentTableName = node.table;
            }
        },
        {
            forEach: "tableConstraint",
            replace(node) {
                if (node.constraintType !== "foreignKey" || knownTables.has(node.referencedTable))
                    return node;
                addAlterTable(node);
            }
        },
        {
            forEach: "column",
            replace(node) {
                if (!node.constraints)
                    return node;
                const found = findColumnForeignKeysWithUnknownReferencedTable(node.constraints, knownTables);
                if (found.length === 0)
                    return node;
                for (const { fk } of found) {
                    const fkName = `${currentTableName}_${node.name}_fk_${fk.referencedTable}`;
                    addAlterTable(fk_column_to_table_constraint_1.foreignKeyColumnToTableConstraint(fk, node.name, fkName));
                }
                const indices = found.map(({ constraintIndex }) => constraintIndex);
                const constraints = modifier_helpers_1.removeSelectedItems(indices, node.constraints);
                const copy = Object.assign({}, node);
                if (!constraints)
                    delete copy.constraints;
                else
                    copy.constraints = constraints;
                return copy;
            }
        },
        {
            forEachChildOf: "orders",
            insertAfter(node) {
                if (node.orderType === "createTable") {
                    const order = deferredFk.get(node.name);
                    if (order)
                        deferredFk.delete(node.name);
                    return order;
                }
            }
        },
    ];
    return {
        astModifiers,
        afterEnd() {
            if (deferredFk.size !== 0) {
                const tableNames = Array.from(deferredFk.keys());
                throw new Error(`Missing referenced table(s): ${tableNames.join(", ")}`);
            }
        }
    };
}
exports.autofixFk = autofixFk;
function findColumnForeignKeysWithUnknownReferencedTable(constraints, knownTables) {
    const result = [];
    if (constraints) {
        for (const [constraintIndex, constraint] of constraints.entries()) {
            if (constraint.constraintType === "foreignKey" && !knownTables.has(constraint.referencedTable)) {
                result.push({
                    constraintIndex,
                    fk: constraint
                });
            }
        }
    }
    return result;
}
