#!/usr/bin/env node
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const commandLineArgs = require("command-line-args");
const commandLineUsage = require("command-line-usage");
const fs_1 = require("fs");
const path_1 = require("path");
const api_1 = require("./api");
const dialect_generators_1 = require("./dialect-generators");
class InvalidArgumentError extends Error {
    constructor(message) {
        super(message);
        this.causeCode = "invalidArgument";
    }
}
const optionDefinitions = [
    {
        name: "help",
        alias: "h",
        type: Boolean,
        description: "Print this help message."
    },
    {
        name: "autofix",
        alias: "a",
        type: Boolean,
        description: "Enable autofix."
    },
    {
        name: "output-dir",
        alias: "o",
        type: String,
        description: "The output directory (optional).",
        typeLabel: "{underline directory}"
    },
    {
        name: "postgresql",
        alias: "p",
        type: Boolean,
        description: "Generate a DDL for Postgresql."
    },
    {
        name: "sqlite",
        alias: "s",
        type: Boolean,
        description: "Generate a DDL for SQLite."
    },
    {
        name: "mariadb",
        alias: "m",
        type: Boolean,
        description: "Generate a DDL for Mariadb or MySQL."
    },
    {
        name: "universal-ddl",
        alias: "u",
        type: Boolean,
        description: "Generate a DDL using the Universal DDL syntax."
    },
    {
        name: "generate-drop",
        alias: "d",
        type: Boolean,
        description: "Generate drop statements (except for the Universal DDL output)."
    },
    {
        name: "encoding",
        alias: "e",
        type: String,
        description: "Encoding for input and output file(s) (default is {underline utf8})."
    },
    {
        name: "force",
        alias: "f",
        type: Boolean,
        description: "Overwrite output files."
    },
    {
        name: "src",
        type: String,
        multiple: true,
        defaultOption: true,
        description: "The source file (by default at last position).",
        typeLabel: "{underline file}",
    },
];
cli();
function cli() {
    const options = parseOptions();
    if (!options)
        return;
    // console.dir(options)
    if (options["help"]) {
        printHelp();
        return;
    }
    try {
        processFiles(options);
    }
    catch (err) {
        if (err.causeCode === "invalidArgument") {
            console.log(`Error: ${err.message}`);
            printHelp();
        }
        else {
            console.error(`Error: ${err.message}`);
        }
    }
}
function printHelp() {
    const sections = [
        {
            header: "Universal DDL",
            content: "Parse DDL scripts in a universal format, then generates DDL scripts for several DBMS."
        },
        {
            header: "Options",
            optionList: optionDefinitions
        }
    ];
    const usage = commandLineUsage(sections);
    console.log(usage);
}
function parseOptions() {
    try {
        return commandLineArgs(optionDefinitions);
    }
    catch (err) {
        console.log(`Error: ${err.message}`);
        printHelp();
    }
}
function processFiles(options) {
    if (!options.src)
        throw new InvalidArgumentError("Missing source file.");
    options.src.map(file => processFile(file, options));
}
function processFile(file, options) {
    let input;
    try {
        input = fs_1.readFileSync(file, { encoding: options.encoding || "utf8" });
    }
    catch (err) {
        throw new InvalidArgumentError(`Cannot read file: ${file}`);
    }
    const bnad = baseNameAndDir(file);
    const ast = api_1.parseDdl(input, { freeze: true, checkConsistency: true, autofix: !!options.autofix });
    if (options.postgresql)
        writeDdl(ast, "postgresql", options, bnad);
    if (options.sqlite)
        writeDdl(ast, "sqlite", options, bnad);
    if (options.mariadb)
        writeDdl(ast, "mariadb", options, bnad);
    if (options["universal-ddl"])
        writeDdl(ast, "universalddl", options, bnad);
}
function writeDdl(ast, dialect, options, bnad) {
    const ddl = dialect_generators_1.generateDdl(ast, dialect, {
        generateDrop: !!options["generate-drop"]
    });
    const dir = normalizePath(options["output-dir"], bnad.directory);
    const file = `${dir}/${bnad.fileBaseName}.${dialect}${bnad.extension}`;
    if (!options.force && fs_1.existsSync(file))
        throw new Error(`Cannot overwrite existing file: ${file}`);
    fs_1.writeFileSync(file, ddl, {
        encoding: options.encoding || "utf8",
    });
}
function normalizePath(path, defaultPath) {
    if (!path)
        return defaultPath;
    path = path.replace(/\/+$/, "");
    if (path)
        return path;
}
function baseNameAndDir(file) {
    const extension = path_1.extname(file);
    return {
        directory: path_1.dirname(file),
        fileBaseName: path_1.basename(file, extension),
        extension
    };
}
