"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ast_helpers_1 = require("../ast-helpers");
class StopChecking extends Error {
    constructor() {
        super(...arguments);
        this.stopCause = "tooManyErrors";
    }
}
class ConsistencyChecker {
    constructor() {
        this.tables = new Map();
        this.errors = [];
    }
    check(ast) {
        try {
            this.checkOrders(ast.orders);
        }
        catch (err) {
            if (err.stopCause !== "tooManyErrors")
                throw err;
        }
    }
    getConsistencyCheckerReport() {
        if (this.errors.length === 0)
            return { valid: true };
        return {
            valid: false,
            errors: [...this.errors]
        };
    }
    checkOrders(orders) {
        for (const order of orders) {
            switch (order.orderType) {
                case "createTable":
                    this.checkCreateTable(order);
                    break;
                case "alterTable":
                    this.checkAlterTable(order);
                    break;
                case "createIndex":
                    this.checkCreateIndex(order);
                    break;
                case "comment":
                    break;
                default:
                    throw new Error(`Unexpected order type: ${order.orderType}`);
            }
        }
    }
    checkCreateTable(order) {
        if (this.tables.has(order.name))
            this.addError(`Duplicated table: ${order.name}`);
        const checkedTable = {
            name: order.name,
            columns: new Map()
        };
        this.tables.set(order.name, checkedTable);
        this.fillCheckedTable(order.entries, checkedTable);
        this.checkTableEntries(order.entries, checkedTable);
    }
    checkAlterTable(order) {
        const checkedTable = this.tables.get(order.table);
        if (!checkedTable)
            this.addError(`Cannot alter table "${order.table}": unknown table`);
        else {
            this.fillCheckedTable(order.add, checkedTable);
            this.checkTableEntries(order.add, checkedTable);
        }
    }
    checkCreateIndex(order) {
        const checkedTable = this.tables.get(order.table);
        if (!checkedTable)
            this.addError(`Cannot create index on table "${order.table}": unknown table`);
        else
            this.findColumns(order.index.columns, checkedTable);
    }
    fillCheckedTable(entries, checkedTable) {
        for (const entry of entries) {
            if (entry.entryType === "column") {
                if (checkedTable.columns.has(entry.name))
                    this.addError(`In table "${checkedTable.name}", duplicated column "${entry.name}"`);
                else {
                    const checkedColumn = {
                        name: entry.name,
                        type: entry.type,
                        typeArgs: entry.typeArgs,
                        columnConstraints: new Set()
                    };
                    checkedTable.columns.set(entry.name, checkedColumn);
                    if (ast_helpers_1.hasColumnConstraint(entry, "primaryKey"))
                        this.setPrimaryKeyOn(checkedTable, [checkedColumn]);
                }
            }
        }
        for (const constraint of entries.filter(({ entryType }) => entryType === "constraint")) {
            if (constraint.constraintType === "primaryKey") {
                const pkColumns = this.findColumns(constraint.columns, checkedTable);
                if (pkColumns)
                    this.setPrimaryKeyOn(checkedTable, pkColumns);
            }
        }
    }
    setPrimaryKeyOn(checkedTable, columns) {
        if (checkedTable.primaryKey)
            this.addError(`In table "${checkedTable.name}", duplicated primary key"`);
        else
            checkedTable.primaryKey = columns;
    }
    checkTableEntries(entries, checkedTable) {
        for (const entry of entries) {
            switch (entry.entryType) {
                case "column":
                    this.checkColumn(entry, checkedTable);
                    break;
                case "constraint":
                    this.checkTableConstraint(entry, checkedTable);
                    break;
                case "comment":
                    break;
                default:
                    throw new Error(`Unexpected type of table entry: ${entry.entryType}`);
            }
        }
    }
    checkColumn(node, checkedTable) {
        const column = checkedTable.columns.get(node.name);
        if (!column)
            throw new Error(`Missing column "${node.name}" in table "${checkedTable.name}"`);
        if (node.constraints) {
            node.constraints.forEach(constraint => this.checkColumnConstraint(constraint, column, checkedTable));
        }
    }
    checkColumnConstraint(node, checkedColumn, checkedTable) {
        switch (node.constraintType) {
            case "notNull":
            case "null":
            case "primaryKey":
            case "unique":
            case "default":
            case "autoincrement":
                if (checkedColumn.columnConstraints.has(node.constraintType)) {
                    this.addError(`In table "${checkedTable.name}", column "${checkedColumn.name}", duplicated` +
                        ` constraint "${node.constraintType}"`);
                }
                else
                    checkedColumn.columnConstraints.add(node.constraintType);
                break;
            case "foreignKey":
                this.checkForeignKey(checkedTable, [checkedColumn], node.referencedTable, node.referencedColumn ? [node.referencedColumn] : undefined);
                break;
            default:
                throw new Error(`Unexpected column constraint: ${node.constraintType}`);
        }
        if (node.constraintType === "default") {
            if (!isValueTypeCompatible(node.value, checkedColumn)) {
                this.addError(`In table "${checkedTable.name}", column "${checkedColumn.name}", default value` +
                    ` "${node.value.value}" is incompatible with type "${checkedColumn.type}"`);
            }
        }
    }
    checkForeignKey(table, columns, refTableName, refColumnNames) {
        if (!columns)
            return;
        const columnNames = columns.map(({ name }) => name);
        const refTable = this.tables.get(refTableName);
        if (!refTable) {
            this.addError(`In table "${table.name}", foreign key "${columnNames.join(`", "`)}":` +
                ` invalid referenced table "${refTableName}"`);
            return;
        }
        if (!refColumnNames)
            refColumnNames = columnNames;
        const refColumns = this.findColumns(refColumnNames, refTable);
        if (!refColumns)
            return;
        if (columns.length !== refColumns.length) {
            this.addError(`In table "${table.name}", foreign key "${columnNames.join(`", "`)}":` +
                ` referenced column(s) don't match: "${refColumnNames.join(`", "`)}"`);
            return;
        }
        for (let i = 0; i < columns.length; ++i) {
            if (!isSameColumnType(columns[i], refColumns[i])) {
                this.addError(`In table "${table.name}", foreign key "${columnNames.join(`", "`)}":` +
                    ` the type of column "${columnNames[i]}" (${toSqlType(columns[i])}) doesn't match with the` +
                    ` type of referenced column "${refColumnNames[i]}" (${toSqlType(refColumns[i])})`);
            }
        }
    }
    checkTableConstraint(node, checkedTable) {
        switch (node.constraintType) {
            case "primaryKey":
            case "unique":
                this.findColumns(node.columns, checkedTable);
                break;
            case "foreignKey":
                const columns = this.findColumns(node.columns, checkedTable);
                if (columns)
                    this.checkForeignKey(checkedTable, columns, node.referencedTable, node.referencedColumns);
                break;
            default:
                throw new Error(`Unexpected table constraint: ${node.constraintType}`);
        }
    }
    findColumns(columnNames, checkedTable) {
        const columns = columnNames.map(columnName => this.findColumn(columnName, checkedTable));
        if (!columns.includes(undefined))
            return columns;
    }
    findColumn(columnName, checkedTable) {
        const column = checkedTable.columns.get(columnName);
        if (!column)
            this.addError(`In table "${checkedTable.name}", unknown column "${columnName}"`);
        return column;
    }
    addError(message) {
        this.errors.push(message);
        if (this.errors.length > 10)
            throw new StopChecking();
    }
}
exports.default = ConsistencyChecker;
function isValueTypeCompatible({ type: valType, value }, { type: colType }) {
    switch (valType) {
        case "sqlExpr":
            if (value === "current_timestamp" || value === "current_time" || value === "current_date")
                return ast_helpers_1.isDataTypeDateOrTime(colType);
            return true;
        case "int":
            return ast_helpers_1.isDataTypeInteger(colType);
        case "float":
            return ast_helpers_1.isDataTypeNumber(colType);
        case "string":
            if (ast_helpers_1.isDataTypeDateOrTime(colType))
                return !isNaN(Date.parse(value));
            return ast_helpers_1.isDataTypeString(colType);
        default:
            throw new Error(`Unexpected value type: ${valType}`);
    }
}
function isSameColumnType(col1, col2) {
    if (col1.type !== col2.type)
        return false;
    if (col1.typeArgs === undefined)
        return col2.typeArgs === undefined;
    if (col2.typeArgs === undefined)
        return false;
    return col1.typeArgs.join(",") === col2.typeArgs.join(",");
}
function toSqlType({ type, typeArgs }) {
    return typeArgs ? `${type}(${typeArgs.join(",")})` : type;
}
