"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function makeGeneratorContext(options, sections, createModifiers) {
    const cx = {
        sections,
        createModifiers,
        options: Object.assign(Object.assign({}, defaultOptions), options),
        gen(sectionName, functionName, ...args) {
            const dict = cx.sections[sectionName];
            if (!dict)
                throw new Error(`Invalid section: ${sectionName}`);
            if (dict[functionName] === undefined)
                throw new Error(`Cannot find "${functionName}" in ${sectionName}`);
            return dict[functionName](cx, ...args);
        }
    };
    return cx;
}
exports.makeGeneratorContext = makeGeneratorContext;
const defaultOptions = {
    indentUnit: "  ",
    generateDrop: false
};
function indentation(cx, level) {
    if (level === 0)
        return "";
    if (level === 1)
        return cx.options.indentUnit;
    const items = [];
    for (let i = 0; i < level; ++i)
        items.push(cx.options.indentUnit);
    return items.join("");
}
exports.indentation = indentation;
function codeToString(cx, code) {
    if (!code)
        return "";
    if (isCodeBlock(code))
        return codeBlockToString(cx, code, 0, "spaced");
    else
        return codeLineToString(code);
}
exports.codeToString = codeToString;
function isCodeBlock(code) {
    return !!(code && code["lines"]);
}
exports.isCodeBlock = isCodeBlock;
function isInlineCode(code) {
    return !!(code && !code["lines"]);
}
exports.isInlineCode = isInlineCode;
function codeBlockToString(cx, block, parentIndent = 0, spaceMode) {
    const { indent, lines, spaceBefore, spaceAfter } = block;
    const curIndent = (indent || 0) + parentIndent;
    const lastIndex = lines.length - 1;
    const content = lines.map((item, index) => {
        if (!item)
            return;
        const prefix = indentation(cx, curIndent);
        if (isCodeBlock(item)) {
            let childSpace;
            if (spaceMode === "spaced")
                childSpace = "cancel";
            else
                childSpace = index === 0 ? "cancelBefore" : index === lastIndex ? "cancelAfter" : undefined;
            return codeBlockToString(cx, item, curIndent, childSpace);
        }
        else
            return prefix + codeLineToString(item);
    }).filter(line => !!line).join(spaceMode === "spaced" ? "\n\n" : "\n");
    if (spaceMode === "cancel")
        return content;
    const prefix = spaceBefore && spaceMode !== "cancelBefore" ? "\n" : "";
    const suffix = spaceAfter && spaceMode !== "cancelAfter" ? "\n" : "";
    return prefix + content + suffix;
}
function codeLineToString({ code, inlineComment }) {
    inlineComment = inlineComment ? `${code ? " " : ""}-- ${inlineComment}` : "";
    return (code || "") + inlineComment;
}
function appendSuffix(items, options) {
    let isLast = true;
    for (let i = items.length - 1; i >= 0; --i) {
        const item = items[i];
        if (isEmptyCodePiece(item, true))
            continue;
        if (isLast) {
            isLast = false;
            if (options.last) {
                appendSuffixToCodePiece(item, options.last);
            }
        }
        else if (options.notLast)
            appendSuffixToCodePiece(item, options.notLast);
    }
    return items;
}
exports.appendSuffix = appendSuffix;
function isEmptyCodePiece(item, ignoreComments = false) {
    if (!item)
        return true;
    if (isCodeBlock(item))
        return !item.lines.find(child => !isEmptyCodePiece(child, ignoreComments));
    else
        return !item.code && (ignoreComments || !item.inlineComment);
}
exports.isEmptyCodePiece = isEmptyCodePiece;
function appendSuffixToCodePiece(item, suffix) {
    if (item) {
        if (isCodeBlock(item))
            appendSuffix(item.lines, { last: suffix });
        else if (item.code)
            item.code += suffix;
    }
}
function tryToMakeInlineBlock(pieces, inlineComment) {
    let single;
    for (const p of pieces) {
        if (isEmptyCodePiece(p))
            continue;
        if (single) {
            return;
        }
        single = p;
    }
    if (!single)
        return;
    if (isCodeBlock(single)) {
        if (single.spaceBefore || single.spaceAfter)
            return;
        return tryToMakeInlineBlock(single.lines, inlineComment);
    }
    if (single.inlineComment)
        inlineComment = inlineComment ? `${single.inlineComment} ${inlineComment}` : single.inlineComment;
    return Object.assign(Object.assign({}, single), { inlineComment });
}
exports.tryToMakeInlineBlock = tryToMakeInlineBlock;
