"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const gen_helpers_1 = require("./gen-helpers");
function makeUniversalDdlGeneratorContext(options) {
    return gen_helpers_1.makeGeneratorContext(options, exports.universalDdlSections);
}
exports.makeUniversalDdlGeneratorContext = makeUniversalDdlGeneratorContext;
const ast = {
    ast(cx, node) {
        return {
            lines: node.orders.map(order => cx.gen("orders", order.orderType, order))
        };
    }
};
const orders = {
    createTable(cx, node) {
        const entries = node.entries.map(entry => cx.gen("tableEntries", entry.entryType, entry));
        let firstInlineCom;
        let lastInlineCom;
        if (node.inlineComment) {
            if (typeof node.inlineComment === "string")
                firstInlineCom = node.inlineComment;
            else if (node.inlineComment.length > 0)
                [firstInlineCom, ...lastInlineCom] = node.inlineComment;
        }
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code: `create table ${node.name} (`,
                    inlineComment: firstInlineCom
                },
                {
                    indent: 1,
                    lines: gen_helpers_1.appendSuffix(entries, { notLast: "," })
                },
                {
                    code: ");",
                    inlineComment: normalizeInlineComment(lastInlineCom)
                },
            ]
        };
    },
    alterTable(cx, node) {
        const inlineComment = normalizeInlineComment(node.inlineComment);
        const entries = node.add.map(entry => cx.gen("tableEntries", entry.entryType, entry));
        const code = `alter table ${node.table} add`;
        const ib = gen_helpers_1.tryToMakeInlineBlock(entries, inlineComment);
        if (ib && ib.code) {
            ib.code = `${code} ${ib.code};`;
            return {
                lines: [toCodeBlockComment(node.blockComment), ib]
            };
        }
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code,
                    inlineComment
                },
                {
                    indent: 1,
                    lines: gen_helpers_1.appendSuffix(entries, { notLast: ",", last: ";" })
                },
            ]
        };
    },
    createIndex(cx, node) {
        const name = node.index.name ? ` ${node.index.name}` : "";
        const columns = node.index.columns.join(", ");
        const unique = node.index["constraintType"] === "unique" ? " unique" : "";
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code: `create${unique} index${name} on ${node.table} (${columns});`,
                    inlineComment: normalizeInlineComment(node.inlineComment)
                },
            ]
        };
    },
    comment(cx, node) {
        return toCodeBlockStandaloneComment(node.blockComment);
    }
};
const tableEntries = {
    column(cx, node) {
        let type = node.type;
        if (node.typeArgs)
            type += `(${node.typeArgs.join(",")})`;
        let constraints = cx.gen("columnChildren", "constraints", node.constraints || []).code;
        constraints = constraints ? ` ${constraints}` : "";
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code: `${node.name} ${type}${constraints}`,
                    inlineComment: normalizeInlineComment(node.inlineComment)
                },
            ]
        };
    },
    constraint(cx, node) {
        return cx.gen("tableConstraints", node.constraintType, node);
    },
    comment(cx, node) {
        return toCodeBlockStandaloneComment(node.blockComment);
    }
};
const columnChildren = {
    constraints(cx, nodes) {
        const constraints = nodes.map(constraint => cx.gen("columnConstraints", constraint.constraintType, constraint).code).join(" ");
        return {
            code: constraints
        };
    }
};
const tableConstraints = {
    primaryKey(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code: `${constraintName}primary key (${node.columns.join(", ")})`,
                    inlineComment: normalizeInlineComment(node.inlineComment)
                },
            ]
        };
    },
    foreignKey(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        const refColumns = node.referencedColumns ? ` (${node.referencedColumns.join(", ")})` : "";
        const del = node.onDelete ? ` on delete ${node.onDelete}` : "";
        const upd = node.onUpdate ? ` on update ${node.onUpdate}` : "";
        const ref = `references ${node.referencedTable}${refColumns}`;
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code: `${constraintName}foreign key (${node.columns.join(", ")}) ${ref}${del}${upd}`,
                    inlineComment: normalizeInlineComment(node.inlineComment)
                },
            ]
        };
    },
    unique(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return {
            lines: [
                toCodeBlockComment(node.blockComment),
                {
                    code: `${constraintName}unique (${node.columns.join(", ")})`,
                    inlineComment: normalizeInlineComment(node.inlineComment)
                },
            ]
        };
    },
};
const columnConstraints = {
    notNull(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return { code: `${constraintName}not null` };
    },
    null(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return { code: `${constraintName}null` };
    },
    default(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return { code: `${constraintName}default ${toSqlValue(node.value)}` };
    },
    primaryKey(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return { code: `${constraintName}primary key` };
    },
    autoincrement(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return { code: `${constraintName}autoincrement` };
    },
    unique(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        return { code: `${constraintName}unique` };
    },
    foreignKey(cx, node) {
        const constraintName = node.name ? `constraint ${node.name} ` : "";
        const del = node.onDelete ? ` on delete ${node.onDelete}` : "";
        const upd = node.onUpdate ? ` on update ${node.onUpdate}` : "";
        const refCol = node.referencedColumn ? ` (${node.referencedColumn})` : "";
        return { code: `${constraintName}references ${node.referencedTable}${refCol}${del}${upd}` };
    },
};
exports.universalDdlSections = {
    ast,
    orders,
    tableEntries,
    columnChildren,
    tableConstraints,
    columnConstraints,
};
function toSqlValue({ type, value }) {
    switch (type) {
        case "float":
        case "int":
            return `${value}`;
        case "string":
            return "'" + value.replace(/'/g, "''") + "'";
        case "sqlExpr":
            return value;
        default:
            throw new Error(`Unexpected value type: ${type}`);
    }
}
exports.toSqlValue = toSqlValue;
function toCodeBlockStandaloneComment(blockComment, indent) {
    return {
        indent,
        lines: blockComment.split("\n").map(inlineComment => ({ inlineComment })),
        spaceBefore: true,
        spaceAfter: true
    };
}
exports.toCodeBlockStandaloneComment = toCodeBlockStandaloneComment;
function toCodeBlockComment(blockComment, indent) {
    if (!blockComment)
        return;
    return {
        indent,
        lines: blockComment.split("\n").map(inlineComment => ({ inlineComment }))
    };
}
exports.toCodeBlockComment = toCodeBlockComment;
function normalizeInlineComment(inlineComment) {
    if (!inlineComment || typeof inlineComment === "string")
        return inlineComment;
    return inlineComment.map(s => s.trim()).join(" ");
}
exports.normalizeInlineComment = normalizeInlineComment;
