"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function createRdsFromAst({ orders }) {
    const astTables = orders.filter(({ orderType }) => orderType === "createTable");
    const astAlterTables = orders.filter(({ orderType }) => orderType === "alterTable");
    const astCreateIndexes = orders.filter(({ orderType }) => orderType === "createIndex");
    const tables = {};
    for (const astTable of astTables)
        tables[astTable.name] = createRdsTable(astTable);
    astAlterTables.forEach(alterTable => addColumnsFromAlterTable(alterTable, tables));
    astTables.forEach(({ name, entries }) => fillTableAndColumnConstraints(name, entries, tables));
    astAlterTables.forEach(({ table, add }) => fillTableAndColumnConstraints(table, add, tables));
    astCreateIndexes.forEach(createIndex => fillTableIndexes(createIndex, tables));
    return { tables };
}
exports.createRdsFromAst = createRdsFromAst;
function createRdsTable({ name, entries, blockComment, inlineComment }) {
    const astColumns = entries.filter(({ entryType }) => entryType === "column");
    const columns = {};
    const table = {
        name,
        columns,
        constraints: {},
        blockComment,
        inlineComment
    };
    for (const column of astColumns)
        columns[column.name] = createRdsColumn(column, table);
    return table;
}
function createRdsColumn({ name, type, typeArgs, blockComment, inlineComment }, table) {
    return {
        table,
        name,
        type,
        typeArgs: typeArgs ? [...typeArgs] : undefined,
        constraints: {},
        blockComment,
        inlineComment
    };
}
function addColumnsFromAlterTable({ table: tableName, add: entries }, tables) {
    const table = getRdsTable(tableName, tables);
    if (!table)
        throw new Error(`Error in "alter table ${tableName}": unknown table`);
    const astColumns = entries.filter(({ entryType }) => entryType === "column");
    for (const column of astColumns)
        table.columns[column.name] = createRdsColumn(column, table);
}
function fillTableAndColumnConstraints(tableName, entries, tables) {
    const table = getRdsTable(tableName, tables);
    fillTableConstraints(entries.filter(({ entryType }) => entryType === "constraint"), table, tables);
    const astColumns = entries.filter(({ entryType }) => entryType === "column");
    astColumns.forEach(({ name, constraints }) => {
        if (constraints)
            fillColumnConstraints(constraints, getRdsColumn(name, table), tables);
    });
}
function fillTableConstraints(astConstraints, table, tables) {
    for (const astConstraint of astConstraints) {
        switch (astConstraint.constraintType) {
            case "primaryKey":
                fillPrimaryKeyConstraint(astConstraint, table);
                break;
            case "unique":
                fillUniqueConstraint(astConstraint, table);
                break;
            case "foreignKey":
                fillForeignKeyConstraint(astConstraint, table, tables);
                break;
            default:
                throw new Error(`Unknown table constraint type: ${astConstraint.constraintType}`);
        }
    }
}
function fillPrimaryKeyConstraint(astConstraint, table) {
    if (table.constraints.primaryKey)
        throw new Error(`Table ${table.name} cannot have several primary keys`);
    const constraint = createPrimaryKeyConstraint(astConstraint, table);
    table.constraints.primaryKey = constraint;
    if (constraint.columns.length === 1)
        constraint.columns[0].constraints.primaryKey = constraint;
}
function createPrimaryKeyConstraint(astConstraint, table) {
    const constraint = {
        constraintType: "primaryKey",
        table,
        columns: astConstraint.columns.map(columnName => getRdsColumn(columnName, table)),
    };
    if (astConstraint.blockComment)
        constraint.blockComment = astConstraint.blockComment;
    if (astConstraint.inlineComment)
        constraint.inlineComment = astConstraint.inlineComment;
    return constraint;
}
function fillUniqueConstraint(astConstraint, table) {
    const constraint = createUniqueConstraint(astConstraint, table);
    if (!table.constraints.uniqueConstraints)
        table.constraints.uniqueConstraints = [];
    table.constraints.uniqueConstraints.push(constraint);
    if (constraint.columns.length === 1)
        constraint.columns[0].constraints.unique = constraint;
}
function createUniqueConstraint(astConstraint, table) {
    const constraint = {
        constraintType: "unique",
        table,
        columns: astConstraint.columns.map(columnName => getRdsColumn(columnName, table)),
    };
    if (astConstraint.blockComment)
        constraint.blockComment = astConstraint.blockComment;
    if (astConstraint.inlineComment)
        constraint.inlineComment = astConstraint.inlineComment;
    return constraint;
}
function fillForeignKeyConstraint(astConstraint, table, tables) {
    const constraint = createForeignKeyConstraint(astConstraint, table, tables);
    if (!table.constraints.foreignKeys)
        table.constraints.foreignKeys = [];
    table.constraints.foreignKeys.push(constraint);
    if (!constraint.referencedTable.referencedBy)
        constraint.referencedTable.referencedBy = [];
    constraint.referencedTable.referencedBy.push(constraint);
    if (constraint.columns.length === 1) {
        const column = constraint.columns[0];
        if (!column.constraints.references)
            column.constraints.references = [];
        column.constraints.references.push(constraint);
        const referencedColumn = constraint.referencedColumns[0];
        if (!referencedColumn.referencedBy)
            referencedColumn.referencedBy = [];
        referencedColumn.referencedBy.push(constraint);
    }
}
function createForeignKeyConstraint(astConstraint, table, tables) {
    const referencedTable = getRdsTable(astConstraint.referencedTable, tables);
    const constraint = {
        constraintType: "foreignKey",
        table,
        columns: astConstraint.columns.map(columnName => getRdsColumn(columnName, table)),
        referencedTable,
        referencedColumns: getRdsColumns(astConstraint.referencedColumns || astConstraint.columns, referencedTable)
    };
    if (astConstraint.onDelete)
        constraint.onDelete = astConstraint.onDelete;
    if (astConstraint.onUpdate)
        constraint.onUpdate = astConstraint.onUpdate;
    if (astConstraint.blockComment)
        constraint.blockComment = astConstraint.blockComment;
    if (astConstraint.inlineComment)
        constraint.inlineComment = astConstraint.inlineComment;
    return constraint;
}
function fillColumnConstraints(astConstraints, column, tables) {
    for (const astConstraint of astConstraints) {
        switch (astConstraint.constraintType) {
            case "notNull":
                column.constraints.notNull = true;
                break;
            case "null":
                break;
            case "primaryKey":
                fillPrimaryKeyConstraint(toAstPrimaryKeyTableConstraint(astConstraint, column), column.table);
                break;
            case "autoincrement":
                column.constraints.autoincrement = true;
                break;
            case "unique":
                fillUniqueConstraint(toAstUniqueTableConstraint(astConstraint, column), column.table);
                break;
            case "foreignKey":
                fillForeignKeyConstraint(toAstForeignKeyTableConstraint(astConstraint, column), column.table, tables);
                break;
            case "default":
                column.constraints.default = {
                    value: astConstraint.value.value
                };
                if (astConstraint.value.type === "sqlExpr")
                    column.constraints.default.isSqlExpr = true;
                break;
            default:
                throw new Error(`Unknown column constraint type: ${astConstraint.constraintType}`);
        }
    }
}
function toAstPrimaryKeyTableConstraint(astSource, column) {
    const constraint = {
        entryType: "constraint",
        constraintType: "primaryKey",
        columns: [column.name],
    };
    if (astSource.name)
        constraint.name = astSource.name;
    return constraint;
}
function toAstUniqueTableConstraint(astSource, column) {
    const constraint = {
        entryType: "constraint",
        constraintType: "unique",
        columns: [column.name],
    };
    if (astSource.name)
        constraint.name = astSource.name;
    return constraint;
}
function toAstForeignKeyTableConstraint(astSource, column) {
    const constraint = {
        entryType: "constraint",
        constraintType: "foreignKey",
        columns: [column.name],
        referencedTable: astSource.referencedTable,
        referencedColumns: [column.name],
    };
    if (astSource.name)
        constraint.name = astSource.name;
    if (astSource.onDelete)
        constraint.onDelete = astSource.onDelete;
    if (astSource.onUpdate)
        constraint.onUpdate = astSource.onUpdate;
    return constraint;
}
function fillTableIndexes({ table: tableName, index }, tables) {
    const table = getRdsTable(tableName, tables);
    if (isAstUniqueTableConstraint(index))
        fillUniqueConstraint(index, table);
    else
        fillIndex(index, table);
}
function isAstUniqueTableConstraint(index) {
    return index["constraintType"] === "unique";
}
function fillIndex(astIndex, table) {
    const index = {
        table,
        columns: astIndex.columns.map(columnName => getRdsColumn(columnName, table)),
    };
    if (astIndex.blockComment)
        index.blockComment = astIndex.blockComment;
    if (astIndex.inlineComment)
        index.inlineComment = astIndex.inlineComment;
    if (!table.indexes)
        table.indexes = [];
    table.indexes.push(index);
}
function getRdsTable(tableName, tables) {
    const table = tables[tableName];
    if (!table)
        throw new Error(`Unknown table "${tableName}"`);
    return table;
}
function getRdsColumn(columnName, table) {
    const column = table.columns[columnName];
    if (!column)
        throw new Error(`Unknown column "${columnName}" in table "${table.name}"`);
    return column;
}
function getRdsColumns(columnNames, table) {
    return columnNames.map(columnName => getRdsColumn(columnName, table));
}
